/* gmos - GTK Metal/Oxyde Semiconductor
 * Copyright (C) 1998 Emmanuel Rouat
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <math.h>
#include <nan.h>
#include "gtkmain.h"
#include "defines.h"
#include "level1.h"
#include "parsing.h"

double level1(void) 
{
    /* This is the classical level 1 model, AKA Schichman-Hodges model.
     * It is in fact the model that derives quite immediatly from
     * simple hand calculations - it is of course far to simple for
     * serious modelisation 
     *
     * Equations and model parameters conform to the description of
     * the level1 model in 'Mosfet Modeling With Spice' By Daniel Foty
     */



    extern double w,l,vg,vd,vs,vb,Tc;
    extern double gm,gds,vth,vp,vdsat;
    extern int type,region,sb_tied; 
    
    extern struct level1_tech_data n1;
    extern struct level1_tech_data p1;    
    struct  level1_tech_data t1;

    double cox,Phif,gamma,kp,Ut,weff,leff,T;
    double vds=0,vgs=0,vbs=0;
    double nsub,u0,lambda,beta,vt0,vfb,EgT;
    double i=0.0; /* beware:local variable */
  
  
    T=Tc+T0; 
    Ut=k*T/q;	    
   
    vp=N_A; /* Irrelevent parameter in level1 */
	    
    if(sb_tied==TRUE) vb = vs; /* source and bulk tied ? */

    /* Two cases: nmos (default) and pmos */
    
    switch(type){

    case NMOS :
	/* all voltages are refered to the source voltage */
	
	vbs=(vb)-(vs);
	vgs=(vg)-(vs);
	vds=(vd)-(vs);
	
	t1=n1; /* model parameters are nmos */
	break;
	
    case PMOS :
	
	vbs=(vs)-(vb);
	vgs=(vs)-(vg);
	vds=(vs)-(vd);
	
	t1=p1; /* model parameters are pmos */	
	break;
    }

    weff=(w*1e-6);  /* Remember w and l are in microns */
    leff=(l*1e-6)-2*t1.XJ; 
    
    u0=t1.U0*1e-4;
    nsub=t1.NSUB*1e6;  /* units conversion */
        
    
    
    /* some fundamental parameters */
    
    cox=(E0*Eox)/(t1.TOX);	
    kp=u0*cox;	
    Phif=Ut*(log(nsub/ni)); /* Fermi potential */	
    gamma=sqrt(2*E0*Esi*q*nsub)/cox;	
    lambda=(t1.LAMBDA_defined==TRUE) ? t1.LAMBDA : 0;
    
    
    /* now the threshold voltages: */
    
    EgT=1.16-(7.02e-4*T*T/(T+1108)); /* Eg(T) */	
    vfb=(t1.TPG=0) ? (EgT-2*Phif)/2 : 0-(2*Phif+t1.TPG*EgT)/2; /* flatband */ 
    vt0=(t1.VT0_defined==TRUE) ? t1.VT0 : vfb+2*Phif+gamma*sqrt(2*Phif);
    
    vth=vt0+gamma*(sqrt(2*Phif-vbs) - sqrt(2*Phif));
	
    /* saturation voltage */
    
    vdsat=vgs-vth;  /* this is really VDSsat */
    
    
    region=(vds<vdsat) ? LINEAR : SATURATED;
    
    if(vgs<=vth) region=BLOQUED; 
    
    
    beta=kp*(weff/leff);
    
    switch(region) { 
    case LINEAR : 
	i=beta*(vgs-vth-vds/2)*vds*(1+lambda*vds); 
	gm=beta*vds*(1+lambda*vds); 
	gds=beta*((vgs-vth-vds)+lambda*vds*(2*(vgs-vth)-1.5*vds));
	break; 
    case SATURATED : 
	i=beta*((vgs-vth)*(vgs-vth)/2)*(1+lambda*vds); 
	gm=2*i/(vgs-vth); 
	gds=beta*lambda*(vgs-vth)*(vgs-vth)/2;
	break; 
    default:
	i=0; 
	gm=0;
	gds=0;
	break;
    }	    	   




#if DEBUG
    
    printf("beta=%g\nkp=%g\nGamma=%g\nPhif=%g\nlambda=%g\n",beta,kp,gamma,Phif,lambda); 
    
#endif
 

         
    return i ;
    

}



void parse_level1(char * table[MAXLINES], int n)
{
    char s[255];
    char buffer[255],mos_type[]="NMOS";
   
    char seps[]=" +=\t\n";
    char seps2[]=" =\t\n";
    char *token;
    char *variable;
    
    double value;
    
    int i,foo,type,verdict;
    
    struct level1_tech_data tmp={ 
	1,	TRUE,      /* tpg */
	0, 	FALSE,     /* tox */
	0,  	FALSE,     /* nsub */
	0, 	FALSE,     /* xj */
	0, 	FALSE,     /* u0 */
	0,	FALSE,     /* vt0 */
	0,	FALSE,     /* lambda */ };  
    

    
    extern struct level1_tech_data n1;
    extern struct level1_tech_data p1;

    /* first arg is mos type : */
    type=(strcmp(table[0],mos_type)==0) ? NMOS : PMOS;

#if DEBUG
    sprintf(buffer,"Received: type: %d\n",type); 
    message(buffer);
#endif


    for(i=1;i<n;i++){

	strcpy(s,table[i]);
		
	/* 
	 * ok , so we get a string of tokens - lets separate them
	 * We should have an alternance name/value 
	 */
	
	token=strtok(s,seps); 
	while(token != NULL){
		
	    variable=token;		    
	    token=strtok(NULL,seps2);
	    value=atof(token); 

#if DEBUG
	    sprintf(buffer,"Received: %s = %g\n",variable,value); 
	    message(buffer);
#endif

	    foo=affect(variable);
	    
	    switch(foo){
	    case PAS_GLOP :sprintf(buffer,"WARNING:unknown variable %s\n",variable);
		message(buffer);
		break;
	    case GMOS_VT0 :tmp.VT0=value;
		tmp.VT0_defined=TRUE;
		break;
	    case GMOS_LAMBDA :tmp.LAMBDA=value;
		tmp.LAMBDA_defined=TRUE;
		break;
	    case GMOS_U0 :tmp.U0=value;
		tmp.U0_defined=TRUE;
		break;
	    case GMOS_NSUB :tmp.NSUB=value;
		tmp.NSUB_defined=TRUE;
		break;
	    case GMOS_TPG :tmp.TPG=value;
		tmp.TPG_defined=TRUE;
		break;
	    case GMOS_TOX :tmp.TOX=value;
		tmp.TOX_defined=TRUE;
		break;
	    case GMOS_XJ  :tmp.XJ=value;
		tmp.XJ_defined=TRUE;
		break;
	    default :sprintf(buffer,"WARNING:irrelevant variable %s\n",variable);
		message(buffer);
	    }
		        
	    token=strtok(NULL,seps2);
	    
	}
	
	    
    }
    
    /* here we should do some sort of checking to see
     * if we have all variables that we need ...
     */
    
    verdict=check_level1(tmp);
    
    if(verdict==GLOP){	    
	if(type==NMOS) n1=tmp;
	if(type==PMOS) p1=tmp;
    }else{
	switch(verdict){
	case GMOS_NSUB : 
	    message("ERROR: NSUB not defined\n");
	    break;
	case GMOS_U0 : 
	    message("ERROR: U0 not defined\n");
	    break;
	case GMOS_TOX : 
	    message("ERROR: TOX not defined\n");
	    break;    
	    
	}
	
    }
    
}


int check_level1(struct level1_tech_data l1)
{

    /* 
     * here we check the bits that are defined - we stop checking every
     * time that we find an error 
     */


    /* first, we need NSUB and U0 in all cases  AKAIK? */

    if(l1.NSUB_defined==FALSE) return(GMOS_NSUB);
    if(l1.U0_defined==FALSE) return(GMOS_U0);
    
 
    /* then we need TOX - some models make it possible to superseed
     * the value of TOX with COX, but not in this case ? 
     */

    if(l1.TOX_defined==FALSE) return(GMOS_TOX);


    return(GLOP);  /* ok */

}

